import re

from collections import namedtuple


class Advertisement(object):
    """Advertisement data record parser

    >>> ad = Advertisement([2, 1, 0x6, 6, 255, 82, 83, 95, 82, 48])
    >>> for data in ad:
    ...     data
    AD(ad_type=1, data=bytearray(b'\\x06'))
    AD(ad_type=255, data=bytearray(b'RS_R0'))
    >>> list(ad)[0].ad_type == Advertisement.ad_types.flags
    True
    """

    AD = namedtuple("AD", ['ad_type', 'data'])

    class ad_types:
        """
        Assigned numbers for some of `advertisement data types
        <https://www.bluetooth.com/specifications/assigned-numbers/generic-access-profile/>`_.

        flags : "Flags" (0x01)

        complete_local_name : "Complete Local Name" (0x09)

        service_data : "Service Data" (0x16)

        manufacturer_specific_data : "Manufacturer Specific Data" (0xff)
        """
        flags = 0x01
        complete_local_name = 0x09
        service_data = 0x16
        manufacturer_specific_data = 0xff

    def __init__(self, data):
        self.data = data

    def __iter__(self):
        return Advertisement.parse(self.data)

    @classmethod
    def parse(cls, data):
        pos = 0
        while pos < len(data):
            length = data[pos]
            if length < 2:
                return
            try:
                ad_type = data[pos + 1]
            except IndexError:
                return
            next_pos = pos + length + 1
            if ad_type:
                segment = slice(pos + 2, next_pos)
                yield Advertisement.AD(ad_type, bytearray(data[segment]))
            pos = next_pos


class Services(dict):
    """Services dict

    >>> services = Services({'service0': {'c1-aa': 0, 'aa-c2-aa': 1},
    ...                      'service1': {'bb-c3-bb': 2}})
    >>> services.search('c3')
    2
    >>> services.search('c4')
    """

    def search(self, pattern, flags=re.IGNORECASE):
        """Search for characteristic by pattern

        :param pattern: regexp pattern
        :param flags: regexp flags, re.IGNORECASE by default
        """
        for characteristics in self.values():
            for uuid, characteristic in characteristics.items():
                if re.search(pattern, uuid, flags):
                    return characteristic



from able import GATT_SUCCESS
from able.structures import Advertisement, Services
from jnius import PythonJavaClass, java_method
from kivy.logger import Logger


class PythonBluetooth(PythonJavaClass):
    __javainterfaces__ = ['org.able.PythonBluetooth']
    __javacontext__ = 'app'

    def __init__(self, dispatcher):
        super(PythonBluetooth, self).__init__()
        self.dispatcher = dispatcher

    @java_method('(Ljava/lang/String;)V')
    def on_error(self, msg):
        Logger.debug("on_error")
        self.dispatcher.dispatch('on_error', msg)

    @java_method('(Landroid/bluetooth/BluetoothDevice;I[B)V')
    def on_device(self, device, rssi, record):
        self.dispatcher.dispatch('on_device', device, rssi,
                                 Advertisement(record))

    @java_method('(Z)V')
    def on_scan_started(self, success):
        Logger.debug("on_scan_started")
        self.dispatcher.dispatch('on_scan_started', success)

    @java_method('()V')
    def on_scan_completed(self):
        Logger.debug("on_scan_completed")
        self.dispatcher.dispatch('on_scan_completed')

    @java_method('(II)V')
    def on_connection_state_change(self, status, state):
        Logger.debug("on_connection_state_change status={} state: {}".format(
            status, state))
        self.dispatcher.dispatch('on_connection_state_change', status, state)

    @java_method('(ILjava/util/List;)V')
    def on_services(self, status, services):
        services_dict = Services()
        if status == GATT_SUCCESS:
            for service in services.toArray():
                service_uuid = service.getUuid().toString()
                Logger.debug("Service discovered: {}".format(service_uuid))
                services_dict[service_uuid] = {}
                for c in service.getCharacteristics().toArray():
                    characteristic_uuid = c.getUuid().toString()
                    Logger.debug("Characteristic discovered: {}".format(
                        characteristic_uuid))
                    services_dict[service_uuid][characteristic_uuid] = c
        self.dispatcher.dispatch('on_services', status, services_dict)

    @java_method('(Landroid/bluetooth/BluetoothGattCharacteristic;)V')
    def on_characteristic_changed(self, characteristic):
        # uuid = characteristic.getUuid().toString()
        self.dispatcher.dispatch('on_characteristic_changed', characteristic)

    @java_method('(Landroid/bluetooth/BluetoothGattCharacteristic;I)V')
    def on_characteristic_read(self, characteristic, status):
        self.dispatcher.dispatch('on_gatt_release')
        # uuid = characteristic.getUuid().toString()
        self.dispatcher.dispatch('on_characteristic_read',
                                 characteristic,
                                 status)

    @java_method('(Landroid/bluetooth/BluetoothGattCharacteristic;I)V')
    def on_characteristic_write(self, characteristic, status):
        self.dispatcher.dispatch('on_gatt_release')
        # uuid = characteristic.getUuid().toString()
        self.dispatcher.dispatch('on_characteristic_write',
                                 characteristic,
                                 status)

    @java_method('(Landroid/bluetooth/BluetoothGattDescriptor;I)V')
    def on_descriptor_read(self, descriptor, status):
        self.dispatcher.dispatch('on_gatt_release')
        # characteristic = descriptor.getCharacteristic()
        # uuid = characteristic.getUuid().toString()
        self.dispatcher.dispatch('on_descriptor_read', descriptor, status)

    @java_method('(Landroid/bluetooth/BluetoothGattDescriptor;I)V')
    def on_descriptor_write(self, descriptor, status):
        self.dispatcher.dispatch('on_gatt_release')
        # characteristic = descriptor.getCharacteristic()
        # uuid = characteristic.getUuid().toString()
        self.dispatcher.dispatch('on_descriptor_write', descriptor, status)

    @java_method('(II)V')
    def on_rssi_updated(self, rssi, status):
        self.dispatcher.dispatch('on_gatt_release')
        self.dispatcher.dispatch('on_rssi_updated', rssi, status)

    @java_method('(II)V')
    def on_mtu_changed(self, mtu, status):
        self.dispatcher.dispatch('on_gatt_release')
        self.dispatcher.dispatch('on_mtu_changed', mtu, status)



from android import activity
from android.permissions import (
    Permission,
    check_permission,
    request_permission,
)
from jnius import autoclass
from kivy.logger import Logger

from able.android.jni import PythonBluetooth
from able.dispatcher import BluetoothDispatcherBase


Activity = autoclass('android.app.Activity')
BLE = autoclass('org.able.BLE')

BluetoothAdapter = autoclass('android.bluetooth.BluetoothAdapter')
BluetoothDevice = autoclass('android.bluetooth.BluetoothDevice')
BluetoothGattDescriptor = autoclass('android.bluetooth.BluetoothGattDescriptor')

ENABLE_NOTIFICATION_VALUE = BluetoothGattDescriptor.ENABLE_NOTIFICATION_VALUE
ENABLE_INDICATION_VALUE = BluetoothGattDescriptor.ENABLE_INDICATION_VALUE
DISABLE_NOTIFICATION_VALUE = BluetoothGattDescriptor.DISABLE_NOTIFICATION_VALUE


class BluetoothDispatcher(BluetoothDispatcherBase):

    @property
    def bonded_devices(self):
        ble_types = (BluetoothDevice.DEVICE_TYPE_LE, BluetoothDevice.DEVICE_TYPE_DUAL)
        return [
            dev for dev in self._ble.mBluetoothAdapter.getBondedDevices().toArray()
            if dev.getType() in ble_types
        ]


    def _set_ble_interface(self):
        self._events_interface = PythonBluetooth(self)
        self._ble = BLE(self._events_interface)
        activity.bind(on_activity_result=self.on_activity_result)

    def _check_runtime_permissions(self):
        # ACCESS_FINE_LOCATION permission is needed to obtain BLE scan results
        return check_permission(Permission.ACCESS_FINE_LOCATION)

    def _request_runtime_permissions(self):
        request_permission(Permission.ACCESS_FINE_LOCATION,
                           self.on_runtime_permissions)

    def connect_by_device_address(self, address: str):
        address = address.upper()
        if not BluetoothAdapter.checkBluetoothAddress(address):
            raise ValueError(f"{address} is not a valid Bluetooth address")
        # remember address to retry connection request when adapter becomes ready
        self._remote_device_address = address
        adapter = self._ble.getAdapter(self.enable_ble_code)
        if adapter:
            self.connect_gatt(adapter.getRemoteDevice(address))

    def enable_notifications(self, characteristic, enable=True, indication=False):
        if not self.gatt.setCharacteristicNotification(characteristic, enable):
            return False

        if not enable:
            # DISABLE_NOTIFICAITON_VALUE is for disabling
            # both notifications and indications
            descriptor_value = DISABLE_NOTIFICATION_VALUE
        elif indication:
            descriptor_value = ENABLE_INDICATION_VALUE
        else:
            descriptor_value = ENABLE_NOTIFICATION_VALUE

        for descriptor in characteristic.getDescriptors().toArray():
            self.write_descriptor(descriptor, descriptor_value)
        return True

    def on_runtime_permissions(self, permissions, grant_results):
        if permissions and all(grant_results):
            self.start_scan()
        else:
            Logger.error(
                'Permissions necessary to obtain scan results are not granted'
            )
            self.dispatch('on_scan_started', False)

    def on_activity_result(self, requestCode, resultCode, intent):
        if requestCode == self.enable_ble_code:
            self.on_bluetooth_enabled(resultCode == Activity.RESULT_OK)

    def on_bluetooth_enabled(self, enabled):
        if enabled:
            if self._remote_device_address:  # connection by MAC address was requested, without scanning
                self.connect_by_device_address(self._remote_device_address)
            else:
                self.start_scan()
        elif not self._remote_device_address:
            self.dispatch('on_scan_started', False)



import os
from os.path import dirname
from setuptools import setup, convert_path
from setuptools.command.install import install


main_ns = {}
with open(convert_path('able/version.py')) as ver_file:
    exec(ver_file.read(), main_ns)

with open(convert_path('README.rst')) as readme_file:
    long_description = readme_file.read()


class InstallRecipe(install):
    """Command to install `able` recipe,
    copies Java files to distribution `javaclass` directory."""

    def run(self):
        if 'ANDROIDAPI' not in os.environ:
            raise Exception(
                'This recipe should not be installed directly, '
                'only with the buildozer tool.'
            )

        # Find Java classes target directory from the environment
        distribution_dir = os.environ['PYTHONPATH'].split(':')[-1]
        distribution_name = distribution_dir.split('/')[-1]
        javaclass_dir = os.path.join(
            dirname(dirname(distribution_dir)),
            'javaclasses'
        )

        if not os.path.exists(javaclass_dir):
            raise Exception(
                'javaclasses directory is not found. '
                'Please report issue to: https://github.com/b3b/able/issues'
            )

        javaclass_dir = os.path.join(javaclass_dir,
                                     distribution_name, 'org', 'able')
        if not os.path.exists(javaclass_dir):
            os.makedirs(javaclass_dir)

        self.distribution.data_files = [
            (javaclass_dir, [
                'able/src/org/able/BLE.java',
                'able/src/org/able/PythonBluetooth.java',
            ]),
        ]

        install.run(self)


setup(
    name='able_recipe',
    version=main_ns['__version__'],
    packages=['able', 'able.android'],
    description='Bluetooth Low Energy for Android',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author='b3b',
    author_email='ash.b3b@gmail.com',
    install_requires=[],
    url='https://github.com/b3b/able',
    project_urls={
        'Changelog': 'https://github.com/b3b/able/blob/master/CHANGELOG.rst',
    },
    # https://pypi.org/classifiers/
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Android',
        'Topic :: System :: Networking',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='android ble bluetooth kivy',
    license='MIT',
    zip_safe=False,
    cmdclass={
        'install': InstallRecipe,
    },
)
